/*
 * Copyright (c) 2007 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *   
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *   
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *  
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *   
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

package sun.jvm.hotspot.utilities.soql;

import java.util.*;
import sun.jvm.hotspot.oops.*;
import org.mozilla.javascript.Scriptable;
import org.mozilla.javascript.ScriptableObject;

/**
   This is JavaScript wrapper for Java List.
*/

public class JSList extends ScriptableObject {
   public JSList(List list, JSJavaFactory fac) {
      this.list = list;
      this.factory = fac;
   }

   public Object get(String name, Scriptable start) {
      if (name.equals("length")) {
         return new Integer(list.size());
      } else {
         return super.get(name, start);
      }
   }

   public Object get(int index, Scriptable start) {
      if (isInRange(index)) {
	  Object item = list.get(index);
	  return wrapObject(item);
      } else {
          return super.get(index, start);
      }
   }

   public String getClassName() {
      return "JSList";
   }

   public final Object getDefaultValue(Class typeHint) {
      if (typeHint == String.class) {
         return toString();
      } else {
	 return super.getDefaultValue(typeHint);
      }
   }

   public Object[] getIds() {
      Object[] superIds = super.getIds();
      final int size = list.size();
      Object[] res = new Object[superIds.length + size];
      for (int i = 0; i < size; i++) {
	  res[i] = new Integer(i);
      }
      System.arraycopy(superIds, 0, res, size, superIds.length);
      return res;
   }

   public boolean has(String name, Scriptable start) {
      if (name.equals("length")) {
         return true; 
      } else {
         return super.has(name, start);
      }
   }

   public boolean has(int index, Scriptable start) {
      if (isInRange(index)) {
         return true;
      } else {
         return super.has(index, start);
      }
   }

   public void put(String name, Scriptable start, Object value) {
      if (! name.equals("length")) {
         super.put(name, start, value);
      }
   }

   public void put(int index, Scriptable start, Object value) {
      if (! isInRange(index)) {
         super.put(index, start, value);
      }
   }

   public String toString() {
       StringBuffer buf = new StringBuffer();
       buf.append('[');
       for (Iterator itr = list.iterator(); itr.hasNext();) {
	   buf.append(wrapObject(itr.next()));
	   if (itr.hasNext()) {
	       buf.append(", ");
	   }
       }
       buf.append(']');
       return buf.toString();
   }

   //-- Internals only below this point
   private boolean isInRange(int index) {
      return index >= 0 && index < list.size();
   }

   private Object wrapObject(Object obj) {
      return factory.newJSJavaWrapper(obj);
   }

   private final List list;
   private final JSJavaFactory factory;
}
