#ifdef USE_PRAGMA_IDENT_HDR
#pragma ident "@(#)mutableNUMASpace.hpp	1.4 07/03/09 17:14:35 JVM"
#endif
/*
 * Copyright (c) 2007 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *   
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *   
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *  
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *   
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

/*
 * The NUMA-aware allocator (MutableNUMASpace) is basically a modification
 * of MutableSpace which preserves interfaces but implements different 
 * functionality. The space is split into chunks for each locality group 
 * (resizing for adaptive size policy is also supported). For each thread
 * the allocations are performed in the chunk corresponding to the home 
 * locality group of the thread. When any chunk fills-in the young generation
 * collection occurs.
 */

class MutableNUMASpace : public MutableSpace {
  friend class VMStructs;

  class LGRPSpace : public CHeapObj {
     int _lgrp_id;
     MutableSpace* _space;
   public:
     LGRPSpace(int l, MutableSpace* s) : _lgrp_id(l), _space(s) { }
     int lgrp_id() const         { return _lgrp_id; }
     MutableSpace* space() const { return _space;   }
     static bool equal(void* lgrp_id_value, LGRPSpace* p) { 
       return *(int*)lgrp_id_value == p->lgrp_id();
     }
  };
  GrowableArray<LGRPSpace*>* _lgrp_spaces;
  // Check if the NUMA topology has changed. Add and remove spaces if needed.
  // The update can be forced by setting the force parameter equal to true.
  bool update_layout(bool force);
  // Bias region towards the first-touching lgrp. The page_size parameter specifies
  // the maximum used page size.
  void bias_region(MemRegion mr, size_t page_size);
 public:
  GrowableArray<LGRPSpace*>* lgrp_spaces() const { return _lgrp_spaces; }
  MutableNUMASpace();
  virtual ~MutableNUMASpace();
  // Initialization
  virtual void initialize(MemRegion mr, bool clear_space);
  // Update space layout if necessary.
  virtual void update();
  virtual void clear();

  virtual void mangle_unused_area();
  virtual void ensure_parseability();
  virtual size_t used_in_words() const;
  virtual size_t free_in_words() const;

  // Allocation (return NULL if full)
  virtual HeapWord* allocate(size_t word_size);
  virtual HeapWord* cas_allocate(size_t word_size);

  // Debugging
  virtual void print_on(outputStream* st) const;
  virtual void print_short_on(outputStream* st) const;
  virtual void verify(bool allow_dirty) const;

  virtual void set_top(HeapWord* value);
};
