#ifdef USE_PRAGMA_IDENT_SRC
#pragma ident "@(#)vmSymbols.cpp	1.25 07/03/09 17:15:06 JVM"
#endif
/*
 * Copyright (c) 2007 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *   
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *   
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *  
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *   
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

# include "incls/_precompiled.incl"
# include "incls/_vmSymbols.cpp.incl"


symbolOop vmSymbols::_symbols[vmSymbols::vm_symbols_terminating_enum];

symbolOop vmSymbols::_type_signatures[T_VOID+1] = { NULL /*, NULL...*/ };

#define VM_SYMBOL_INITIALIZE(name, string) { \
  symbolOop sym = oopFactory::new_symbol(string, CHECK);  \
  _symbols[VM_SYMBOL_ENUM_NAME(name)] = sym; }


void vmSymbols::initialize(TRAPS) {
  if (!UseSharedSpaces) {
    VM_SYMBOLS_DO(VM_SYMBOL_INITIALIZE)

      _type_signatures[T_BYTE]    = byte_signature();
    _type_signatures[T_CHAR]    = char_signature();
    _type_signatures[T_DOUBLE]  = double_signature();
    _type_signatures[T_FLOAT]   = float_signature();
    _type_signatures[T_INT]     = int_signature();
    _type_signatures[T_LONG]    = long_signature();
    _type_signatures[T_SHORT]   = short_signature();
    _type_signatures[T_BOOLEAN] = bool_signature();
    _type_signatures[T_VOID]    = void_signature();
    // no single signatures for T_OBJECT or T_ARRAY
  }
}


void vmSymbols::oops_do(OopClosure* f, bool do_all) {
  for (int index = 0; index < vm_symbols_terminating_enum; index++) {
    f->do_oop((oop*) &_symbols[index]);
  }
  for (int i = 0; i < T_VOID+1; i++) {
    if (_type_signatures[i] != NULL) {
      assert(i >= T_BOOLEAN, "checking");
      f->do_oop((oop*)&_type_signatures[i]);
    } else if (do_all) {
      f->do_oop((oop*)&_type_signatures[i]);
    }
  }
}


BasicType vmSymbols::signature_type(symbolOop s) {
  assert(s != NULL, "checking");
  for (int i = T_BOOLEAN; i < T_VOID+1; i++) {
    if (s == _type_signatures[i]) {
      return (BasicType)i;
    }
  }
  return T_OBJECT;
}


#define VM_INTRINSIC_INITIALIZE(id, klass, name, sig) #id,
static const char* vm_intrinsic_name_table[] = {
  NULL,
  VM_INTRINSICS_DO(VM_INTRINSIC_INITIALIZE)
  NULL
};

const char* vmIntrinsics::name_at(int raw_id) {
  if ((uint)raw_id < (uint)_vm_intrinsics_terminating_enum)
    return vm_intrinsic_name_table[raw_id];
  else
    return NULL;
}
