#ifdef USE_PRAGMA_IDENT_HDR
#pragma ident "@(#)oop.pcgc.inline.hpp	1.13 07/03/09 17:15:17 JVM"
#endif
/*
 * Copyright (c) 2007 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *   
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *   
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *  
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *   
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

inline void oopDesc::update_contents(ParCompactionManager* cm) {
  // The klass field must be updated before anything else
  // can be done.
  DEBUG_ONLY(klassOopDesc* original_klass = klass());

  // Can the option to update and/or copy be moved up in the
  // call chain to avoid calling into here?
 
  if (PSParallelCompact::should_update_klass(klass())) {
    update_header();
    assert(klass()->is_klass(), "Not updated correctly");
  } else {
    assert(klass()->is_klass(), "Not updated");
  }

  Klass* new_klass = blueprint();
  if (!new_klass->oop_is_typeArray()) {
    // It might contain oops beyond the header, so take the virtual call.
    new_klass->oop_update_pointers(cm, this);
  }
  // Else skip it.  The typeArrayKlass in the header never needs scavenging.
}

inline void oopDesc::update_contents(ParCompactionManager* cm,
				     HeapWord* begin_limit,
				     HeapWord* end_limit) {
  // The klass field must be updated before anything else
  // can be done.
  debug_only(klassOopDesc* original_klass = klass());

  update_contents(cm, klass(), begin_limit, end_limit);
}

inline void oopDesc::update_contents(ParCompactionManager* cm,
				     klassOop old_klass,
				     HeapWord* begin_limit,
				     HeapWord* end_limit) {

  klassOop updated_klass = 
    PSParallelCompact::summary_data().calc_new_klass(old_klass);

  // Needs to be boundary aware for the 64 bit case
  // update_header();
  // The klass has moved.  Is the location of the klass
  // within the limits?
  if ((((HeapWord*)&_klass) >= begin_limit) &&
      (((HeapWord*)&_klass) < end_limit)) {
    set_klass(updated_klass);
  }

  Klass* klass = updated_klass->klass_part();
  if (!klass->oop_is_typeArray()) {
    // It might contain oops beyond the header, so take the virtual call.
    klass->oop_update_pointers(cm, this, begin_limit, end_limit);
  }
  // Else skip it.  The typeArrayKlass in the header never needs scavenging.
}

inline void oopDesc::follow_contents(ParCompactionManager* cm) {
  assert (PSParallelCompact::mark_bitmap()->is_marked(this), 
    "should be marked");
  blueprint()->oop_follow_contents(cm, this);
}

