#ifdef USE_PRAGMA_IDENT_HDR
#pragma ident "@(#)jvmpi.inline.hpp	1.17 07/03/09 17:15:50 JVM"
#endif
/*
 * Copyright (c) 2007 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *   
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *   
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *  
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *   
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

inline bool jvmpi::is_event_enabled(jint event_type) {
  return (event_type >= 31
	  ? (enabled() && (_event_flags_array[event_type] == JVMPI_EVENT_ENABLED)) 
	  : (_event_flags & (1 << event_type)) != 0);
}


inline void* jvmpi::calloc(size_t size) {
  void* p = os::malloc(size);
  if (p == NULL) {
    vm_exit_out_of_memory(size, "jvmpi::malloc");
  }
  memset(p, 0, size);
  return p;
}


inline void jvmpi::free(void* ptr) {
  os::free(ptr);
}


inline void jvmpi::enable_event(jint event_type) {
  if (event_type < 31) {
    _event_flags |= 1 << event_type;
  }
  _event_flags_array[event_type] = JVMPI_EVENT_ENABLED;
}


inline void jvmpi::disable_event(jint event_type) {
  if (event_type < 31) {
    _event_flags &= ~(1 << event_type);
  }
  _event_flags_array[event_type] = JVMPI_EVENT_DISABLED;
}


inline bool jvmpi::enabled() {
  return !!(_event_flags & JVMPI_PROFILING_ON);
}


inline bool jvmpi::is_event_supported(jint event_type) {
  return ((event_type <= JVMPI_MAX_EVENT_TYPE_VAL) &&
	  (_event_flags_array[event_type] != JVMPI_EVENT_NOT_SUPPORTED));
}


inline unsigned int* jvmpi::event_flags_array_at_addr(jint event_type) {
  return &_event_flags_array[event_type];
}
